import functools
from typing import TYPE_CHECKING, Any, Callable, Optional, Type, TypeVar, Union, cast


class OpenBASError(Exception):
    def __init__(
        self,
        error_message: Union[str, bytes] = "",
        response_code: Optional[int] = None,
        response_body: Optional[bytes] = None,
    ) -> None:
        Exception.__init__(self, error_message)
        # Http status code
        self.response_code = response_code
        # Full http response
        self.response_body = response_body
        try:
            # if we receive str/bytes we try to convert to unicode/str to have
            # consistent message types (see #616)
            if TYPE_CHECKING:
                assert isinstance(error_message, bytes)
            self.error_message = error_message.decode()
        except Exception:
            if TYPE_CHECKING:
                assert isinstance(error_message, str)
            self.error_message = error_message

    def __str__(self) -> str:
        if self.response_code is not None:
            return f"{self.response_code}: {self.error_message}"
        return f"{self.error_message}"


class OpenBASAuthenticationError(OpenBASError):
    pass


class OpenBASHttpError(OpenBASError):
    pass


class OpenBASParsingError(OpenBASError):
    pass


class RedirectError(OpenBASError):
    pass


class OpenBASHeadError(OpenBASError):
    pass


class OpenBASGetError(OpenBASError):
    pass


class OpenBASUpdateError(OpenBASError):
    pass


class OpenBASListError(OpenBASError):
    pass


class OpenBASCreateError(OpenBASError):
    pass


class ConfigurationError(OpenBASError):
    pass


# For an explanation of how these type-hints work see:
# https://mypy.readthedocs.io/en/stable/generics.html#declaring-decorators
#
# The goal here is that functions which get decorated will retain their types.
__F = TypeVar("__F", bound=Callable[..., Any])


def on_http_error(error: Type[Exception]) -> Callable[[__F], __F]:
    def wrap(f: __F) -> __F:
        @functools.wraps(f)
        def wrapped_f(*args: Any, **kwargs: Any) -> Any:
            try:
                return f(*args, **kwargs)
            except OpenBASHttpError as e:
                raise error(e.error_message, e.response_code, e.response_body) from e

        return cast(__F, wrapped_f)

    return wrap


# Export manually to keep mypy happy
__all__ = [
    "ConfigurationError",
    "OpenBASAuthenticationError",
    "OpenBASHttpError",
    "OpenBASParsingError",
    "RedirectError",
    "OpenBASHeadError",
    "OpenBASListError",
    "OpenBASGetError",
    "OpenBASUpdateError",
]



from typing import Any, Dict

from pyobas import exceptions as exc
from pyobas.apis.inject_expectation.model import (
    DetectionExpectation,
    ExpectationTypeEnum,
    PreventionExpectation,
)
from pyobas.base import RESTManager, RESTObject
from pyobas.mixins import ListMixin, UpdateMixin
from pyobas.utils import RequiredOptional


class InjectExpectation(RESTObject):
    _id_attr = "inject_expectation_id"


class InjectExpectationManager(ListMixin, UpdateMixin, RESTManager):
    _path = "/injects/expectations"
    _obj_cls = InjectExpectation
    _update_attrs = RequiredOptional(required=("collector_id", "result", "is_success"))

    @exc.on_http_error(exc.OpenBASUpdateError)
    def expectations_assets_for_source(
        self, source_id: str, expiration_time: int = None, **kwargs: Any
    ) -> Dict[str, Any]:
        path = f"{self.path}/assets/" + source_id
        result = self.openbas.http_get(
            path,
            query_data=(
                {"expiration_time": expiration_time} if expiration_time else None
            ),
            **kwargs,
        )
        return result

    def expectations_models_for_source(self, source_id: str, **kwargs: Any):
        """Returns all expectations from OpenBAS that have had no result yet
            from the source_id (e.g. collector).

        :param source_id: the identifier of the collector requesting expectations
        :type source_id: str
        :param kwargs: additional data to pass to the endpoint
        :type kwargs: dict, optional

        :return: a list of expectation objects
        :rtype: list[DetectionExpectation|PreventionExpectation]
        """
        # TODO: we should implement a more clever mechanism to obtain
        #   specialised Expectation instances rather than just if/elseing
        #   through this list of possibilities.
        expectations = []
        for expectation_dict in self.expectations_assets_for_source(
            source_id=source_id, **kwargs
        ):
            if (
                expectation_dict["inject_expectation_type"]
                == ExpectationTypeEnum.Detection.value
            ):
                expectations.append(
                    DetectionExpectation(**expectation_dict, api_client=self)
                )
            elif (
                expectation_dict["inject_expectation_type"]
                == ExpectationTypeEnum.Prevention.value
            ):
                expectations.append(
                    PreventionExpectation(**expectation_dict, api_client=self)
                )
            else:
                expectations.append(
                    PreventionExpectation(**expectation_dict, api_client=self)
                )
        return expectations

    @exc.on_http_error(exc.OpenBASUpdateError)
    def prevention_expectations_for_source(
        self, source_id: str, **kwargs: Any
    ) -> Dict[str, Any]:
        path = f"{self.path}/prevention" + source_id
        result = self.openbas.http_get(path, **kwargs)
        return result

    @exc.on_http_error(exc.OpenBASUpdateError)
    def detection_expectations_for_source(
        self, source_id: str, expiration_time: int = None, **kwargs: Any
    ) -> Dict[str, Any]:
        path = f"{self.path}/detection/" + source_id
        result = self.openbas.http_get(
            path,
            query_data=(
                {"expiration_time": expiration_time} if expiration_time else None
            ),
            **kwargs,
        )
        return result

    @exc.on_http_error(exc.OpenBASUpdateError)
    def update(
        self,
        inject_expectation_id: str,
        inject_expectation: Dict[str, Any],
        **kwargs: Any,
    ) -> Dict[str, Any]:
        path = f"{self.path}/{inject_expectation_id}"
        result = self.openbas.http_put(path, post_data=inject_expectation, **kwargs)
        return result

    @exc.on_http_error(exc.OpenBASUpdateError)
    def bulk_update(
        self,
        inject_expectation_input_by_id: Dict[str, Dict[str, Any]],
        **kwargs: Any,
    ) -> None:
        path = f"{self.path}/bulk"
        self.openbas.http_put(
            path, post_data={"inputs": inject_expectation_input_by_id}, **kwargs
        )



import sched
import time

from pyobas.daemons import BaseDaemon
from pyobas.utils import PingAlive

DEFAULT_PERIOD_SECONDS = 60


class CollectorDaemon(BaseDaemon):
    """Implementation of a daemon of Collector type. Note that it requires
    specific configuration keys to run its setup.
    `collector_icon_filepath`: relative path to an icon image (preferably PNG)
    `collector_id`: unique identifier for the collector (UUIDv4)
    `collector_period`: time to wait in seconds between each loop execution; note
    that this time is added to the time the loop takes to run, so the actual total
    time between each loop start is time_of_loop+period.
    """

    def _setup(self):
        if self._configuration.get("collector_period") is None:
            self._configuration.set("collector_period", DEFAULT_PERIOD_SECONDS)
        icon_path = self._configuration.get("collector_icon_filepath")
        icon_name = self._configuration.get("collector_id") + ".png"
        with open(icon_path, "rb") as icon_file_handle:
            collector_icon = (icon_name, icon_file_handle, "image/png")
            document = self.api.document.upsert(document={}, file=collector_icon)
            if self._configuration.get("collector_platform") is not None:
                security_platform = self.api.security_platform.upsert(
                    {
                        "asset_name": self._configuration.get("collector_name"),
                        "asset_external_reference": self._configuration.get(
                            "collector_id"
                        ),
                        "security_platform_type": self._configuration.get(
                            "collector_platform"
                        ),
                        "security_platform_logo_light": document.get("document_id"),
                        "security_platform_logo_dark": document.get("document_id"),
                    }
                )
            else:
                security_platform = {}
            security_platform_id = security_platform.get("asset_id")
            config = {
                "collector_id": self._configuration.get("collector_id"),
                "collector_name": self._configuration.get("collector_name"),
                "collector_type": self._configuration.get("collector_type"),
                "collector_period": self._configuration.get("collector_period"),
                "collector_security_platform": security_platform_id,
            }
        with open(icon_path, "rb") as icon_file_handle:
            collector_icon = (icon_name, icon_file_handle, "image/png")
            self.api.collector.create(config, collector_icon)

        PingAlive(self.api, config, self.logger, "collector").start()

    def _start_loop(self):
        scheduler = sched.scheduler(time.time, time.sleep)
        delay = self._configuration.get("collector_period")
        self._try_callback()
        scheduler.enter(
            delay=delay,
            priority=1,
            action=self.__schedule,
            argument=(scheduler, self._try_callback, delay),
        )
        scheduler.run()

    def __schedule(self, scheduler, callback, delay):
        callback()
        scheduler.enter(
            delay=delay,
            priority=1,
            action=self.__schedule,
            argument=(scheduler, callback, delay),
        )



import os
import os.path
from typing import Any, Dict, Optional

import yaml
from pydantic import BaseModel, Field

from pyobas.configuration.sources import DictionarySource, EnvironmentSource

CONFIGURATION_TYPES = str | int | bool | Any | None


def is_truthy(value: str) -> bool:
    """Asserts whether a given string signals a "True" value

    :param value: value to test
    :type value: str

    :return: whether the string represents True or not.
    :rtype: bool
    """
    return value.lower() in ["yes", "true"]


def is_falsy(value: str) -> bool:
    """Asserts whether a given string signals a "False" value

    :param value: value to test
    :type value: str

    :return: whether the string represents False or not.
    :rtype: bool
    """
    return value.lower() in ["no", "false"]


class ConfigurationHint(BaseModel):
    """An individual configuration hint. This allows for specifying
    where any given configuration key can be found, in env vars,
    config files. Additionally, it may define a default value or
    a discrete override value.
    """

    data: Optional[CONFIGURATION_TYPES] = Field(default=None)
    """Override value; when set, getting the configuration value for
        the key described in this instance returns this value.
    """
    env: Optional[str] = Field(default=None)
    """Defines which env var should be read for getting the value
    """
    file_path: Optional[list[str]] = Field(default=None)
    """Defines a JSON path (nested keys) to follow in the provided
        config file for reaching the value.
        
        Example: ["toplevel", "subkey"] will hint for searching for 
        the config key at { "toplevel": { "subkey": { "config_key"}}
    """
    is_number: Optional[bool] = Field(default=False)
    """Hints at whteher the configuration value should be
        interpreted as a number.
    """
    default: Optional[CONFIGURATION_TYPES] = Field(default=None)
    """When defined, provides a default value for whenever none of the
        hinted locations or the data field have a value.
    """


class Configuration:
    """A configuration object providing ways to an interface for getting
    configuration values. It should be provided with a collection of hints
    to enable its behaviour.

        :param config_hints: a dictionary of hints, for which the key is the
            desired configuration key (e.g. "log_level") and the value is either
            a dictionary of hints (see ConfigurationHint) or a standalone string.
            In the latter case, the string will be interpreted as a default value.

            Example:
            .. code-block:: python
                {
                    "my_config_key": {
                        "env" : "MY_CONFIG_VALUE_ENV_VAR",
                        "file_path": ["first_level", "second_level"]
                    },
                    "my_other_config_key: "discrete value"
                }
        :type config_hints: Dict[str, dict | str]
        :param config_values: dictionary of config values to preemptively load into the
            Configuration object. The format of this dictionary should follow the patterns
            chosen in the file_path property of ConfigurationHint object passed
            as config_hints, defaults to None

            Example:
            .. code-block:: python
                {
                    "first_level": {
                        "second_level": {
                            "my_config_key": "some value"
                        }
                    }
                }
        :type config_values: dict (json), optional
        :param config_file_path: path to the configuration file. The file should
            contain a json structure that matches the format of the config_values param,
            defaults to './config.yml' (relative path).
        :type config_file_path: str
    """

    def __init__(
        self,
        config_hints: Dict[str, dict | str],
        config_values: dict = None,
        config_file_path: str = os.path.join(os.curdir, "config.yml"),
    ):
        self.__config_hints = {
            key: (
                ConfigurationHint(**value)
                if isinstance(value, dict)
                else ConfigurationHint(**{"default": value})
            )
            for key, value in config_hints.items()
        }

        file_contents = (
            yaml.load(open(config_file_path), Loader=yaml.FullLoader)
            if os.path.isfile(config_file_path)
            else {}
        )

        self.__config_values = (config_values or {}) | file_contents

    def get(self, config_key: str) -> CONFIGURATION_TYPES:
        """Gets the value pointed to by the configuration key. If the key is defined
        with actual hints (as opposed to a discrete value), it will use those hints to
        potentially find a value. If the key was not defined as part of the supplied
        config_hints, this will always return None.

        :param config_key: the configuration key to search a value for.
        :type config_key: str

        :return: the value pointed to by the configuration key, or None if not found
        :rtype: CONFIGURATION_TYPES
        """
        config = self.__config_hints.get(config_key)
        if config is None:
            return None

        return self.__process_value_to_type(
            config.data or self.__dig_config_sources_for_key(config), config.is_number
        )

    def set(self, config_key: str, value: CONFIGURATION_TYPES):
        """Sets an arbitrary value in the Configuration object, for
        the supplied configuration key, after which any request for the value
        of that key will return this new value.

        :param config_key: the configuration key to set a value for.
        :type config_key: str
        :param value: the new value to set for the configuration key.
        :type value: CONFIGURATION_TYPES
        """
        if config_key not in self.__config_hints:
            self.__config_hints[config_key] = ConfigurationHint(**{"data": value})
        else:
            self.__config_hints[config_key].data = value

    @staticmethod
    def __process_value_to_type(value: CONFIGURATION_TYPES, is_number_hint: bool):
        if value is None:
            return value
        if isinstance(value, int) or is_number_hint:
            return int(value)
        if isinstance(value, str):
            if is_truthy(value):
                return True
            if is_falsy(value):
                return False
            if len(value) == 0:
                return None
        return value

    def __dig_config_sources_for_key(
        self, config: ConfigurationHint
    ) -> CONFIGURATION_TYPES:
        result = EnvironmentSource.get(config.env) or DictionarySource.get(
            config.file_path, self.__config_values
        )
        return result or config.default



from enum import Enum
from typing import List
from uuid import UUID

from pydantic import BaseModel
from thefuzz import fuzz

from pyobas.signatures.signature_type import SignatureType
from pyobas.signatures.types import MatchTypes, SignatureTypes


class ExpectationTypeEnum(str, Enum):
    """Types of Expectations"""

    Detection = "DETECTION"
    Prevention = "PREVENTION"
    Other = "other"

    @classmethod
    def _missing_(cls, value):
        return cls.Other


class ExpectationSignature(BaseModel):
    """An expectation signature describes a known marker potentially
    found in alerting data in security software. For example, an
    expectation signature can be a process image name, a command
    line, or any other relevant piece of data.
    """

    type: SignatureTypes
    value: str


class Expectation(BaseModel):
    """An expectation represents an expected outcome of a BAS run.
    For example, in the case of running an attack command line, the
    expectation may be that security software has _detected_ it, while
    another expectation may be that the attack was _prevented_.
    """

    inject_expectation_id: UUID
    inject_expectation_signatures: List[ExpectationSignature]

    success_label: str = "Success"
    failure_label: str = "Failure"

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.__api_client = kw["api_client"]

    def update(self, success, sender_id, metadata):
        """Update the expectation object in OpenBAS with the supplied outcome.

        :param success: whether the expectation was fulfilled (true) or not (false)
        :type success: bool
        :param sender_id: identifier of the collector that is updating the expectation
        :type sender_id: string
        :param metadata: arbitrary dictionary of additional data relevant to updating the expectation
        :type metadata: dict[string,string]
        """
        self.__api_client.update(
            self.inject_expectation_id,
            inject_expectation={
                "collector_id": sender_id,
                "result": (self.success_label if success else self.failure_label),
                "is_success": success,
                "metadata": metadata,
            },
        )

    def match_alert(self, relevant_signature_types: list[SignatureType], alert_data):
        """Matches an alert's data against the current expectation signatures
        to see if the alert is relevant to the current expectation's inject,
        i.e. this alert was triggered by the execution of the inject to which
        belongs the expectation.

        :param relevant_signature_types: filter of signature types that we want to consider.
            Only the signature types listed in this collection may be checked for matching.
        :type relevant_signature_types: list[SignatureType]
        :param alert_data: list of possibly relevant markers found in an alert.
        :type alert_data: dict[SignatureTypes, dict]

        :return: whether the alert matches the expectation signatures or not.
        :rtype: bool
        """
        relevant_expectation_signatures = [
            signature
            for signature in self.inject_expectation_signatures
            if signature.type in [type.label for type in relevant_signature_types]
        ]
        if not any(relevant_expectation_signatures):
            return False

        for relevant_expectation_signature in relevant_expectation_signatures:
            if not (
                alert_signature_for_type := alert_data.get(
                    relevant_expectation_signature.type.value
                )
            ):
                return False

            if alert_signature_for_type[
                "type"
            ] == MatchTypes.MATCH_TYPE_FUZZY and not self.match_fuzzy(
                alert_signature_for_type["data"],
                relevant_expectation_signature.value,
                alert_signature_for_type["score"],
            ):
                return False
            if alert_signature_for_type[
                "type"
            ] == MatchTypes.MATCH_TYPE_SIMPLE and not self.match_simple(
                alert_signature_for_type["data"], relevant_expectation_signature.value
            ):
                return False

        return True

    @staticmethod
    def match_fuzzy(tested: list[str], reference: str, threshold: int):
        """Applies a fuzzy match against a known reference to a list of candidates

        :param tested: list of strings candidate for fuzzy matching
        :type tested: list[str]
        :param reference: the reference against which to try to fuzzy match
        :type reference: str
        :param threshold: string overlap percentage threshold above which to declare a match
        :type threshold: int

        :return: whether any of the candidate is a match against the reference
        :rtype: bool
        """
        actual_tested = [tested] if isinstance(tested, str) else tested
        for value in actual_tested:
            ratio = fuzz.ratio(value, reference)
            if ratio >= threshold:
                return True
        return False

    @staticmethod
    def match_simple(tested: list[str], reference: str):
        """A simple strict, case-sensitive string matching between a list of
            candidates and a reference.

        :param tested: list of strings candidate for fuzzy matching
        :type tested: list[str]
        :param reference: the reference against which to try to fuzzy match
        :type reference: str

        :return: whether any of the candidate is a match against the reference
        :rtype: bool
        """
        return Expectation.match_fuzzy(tested, reference, threshold=100)


class DetectionExpectation(Expectation):
    """An expectation that is specific to Detection, i.e. that is used
    by OpenBAS to assert that an inject's execution was detected.
    """

    success_label: str = "Detected"
    failure_label: str = "Not Detected"


class PreventionExpectation(Expectation):
    """An expectation that is specific to Prevention, i.e. that is used
    by OpenBAS to assert that an inject's execution was prevented.
    """

    success_label: str = "Prevented"
    failure_label: str = "Not Prevented"



from typing import Any, cast

from pyobas.base import RESTManager, RESTObject
from pyobas.mixins import GetWithoutIdMixin, UpdateMixin


class Me(RESTObject):
    _id_attr = None
    _repr_attr = "user_email"


class MeManager(GetWithoutIdMixin, UpdateMixin, RESTManager):
    _path = "/me"
    _obj_cls = Me

    def get(self, **kwargs: Any) -> Me:
        return cast(Me, super().get(**kwargs))



import base64
import json
import os
import os.path
import re
import sched
import ssl
import tempfile
import threading
import time
import traceback
from typing import Callable, Dict, List

import pika
from thefuzz import fuzz

from pyobas import OpenBAS, utils
from pyobas.configuration import Configuration
from pyobas.daemons import CollectorDaemon
from pyobas.exceptions import ConfigurationError

TRUTHY: List[str] = ["yes", "true", "True"]
FALSY: List[str] = ["no", "false", "False"]


# As cert must be written in files to be loaded in ssl context
# Creates a temporary file in the most secure manner possible
def data_to_temp_file(data):
    # The file is readable and writable only by the creating user ID.
    # If the operating system uses permission bits to indicate whether a
    # file is executable, the file is executable by no one. The file
    # descriptor is not inherited by children of this process.
    file_descriptor, file_path = tempfile.mkstemp()
    with os.fdopen(file_descriptor, "w") as open_file:
        open_file.write(data)
        open_file.close()
    return file_path


def is_memory_certificate(certificate):
    return certificate.startswith("-----BEGIN")


def ssl_cert_chain(ssl_context, cert_data, key_data, passphrase):
    if cert_data is None:
        return

    cert_file_path = None
    key_file_path = None

    # Cert loading
    if cert_data is not None and is_memory_certificate(cert_data):
        cert_file_path = data_to_temp_file(cert_data)
    cert = cert_file_path if cert_file_path is not None else cert_data

    # Key loading
    if key_data is not None and is_memory_certificate(key_data):
        key_file_path = data_to_temp_file(key_data)
    key = key_file_path if key_file_path is not None else key_data

    # Load cert
    ssl_context.load_cert_chain(cert, key, passphrase)
    # Remove temp files
    if cert_file_path is not None:
        os.unlink(cert_file_path)
    if key_file_path is not None:
        os.unlink(key_file_path)


def ssl_verify_locations(ssl_context, certdata):
    if certdata is None:
        return

    if is_memory_certificate(certdata):
        ssl_context.load_verify_locations(cadata=certdata)
    else:
        ssl_context.load_verify_locations(cafile=certdata)


def create_mq_ssl_context(config) -> ssl.SSLContext:
    config_obj = Configuration(
        config_hints={
            "MQ_USE_SSL_CA": {
                "env": "MQ_USE_SSL_CA",
                "file_path": ["mq", "use_ssl_ca"],
            },
            "MQ_USE_SSL_CERT": {
                "env": "MQ_USE_SSL_CERT",
                "file_path": ["mq", "use_ssl_cert"],
            },
            "MQ_USE_SSL_KEY": {
                "env": "MQ_USE_SSL_KEY",
                "file_path": ["mq", "use_ssl_key"],
            },
            "MQ_USE_SSL_REJECT_UNAUTHORIZED": {
                "env": "MQ_USE_SSL_REJECT_UNAUTHORIZED",
                "file_path": ["mq", "use_ssl_reject_unauthorized"],
                "is_number": False,
                "default": False,
            },
            "MQ_USE_SSL_PASSPHRASE": {
                "env": "MQ_USE_SSL_PASSPHRASE",
                "file_path": ["mq", "use_ssl_passphrase"],
            },
        },
        config_values=config,
    )
    use_ssl_ca = config_obj.get("MQ_USE_SSL_CA")
    use_ssl_cert = config_obj.get("MQ_USE_SSL_CERT")
    use_ssl_key = config_obj.get("MQ_USE_SSL_KEY")
    use_ssl_reject_unauthorized = config_obj.get("MQ_USE_SSL_REJECT_UNAUTHORIZED")
    use_ssl_passphrase = config_obj.get("MQ_USE_SSL_PASSPHRASE")
    ssl_context = ssl.create_default_context()
    # If no rejection allowed, use private function to generate unverified context
    if not use_ssl_reject_unauthorized:
        # noinspection PyUnresolvedReferences,PyProtectedMember
        ssl_context = ssl._create_unverified_context()
    ssl_verify_locations(ssl_context, use_ssl_ca)
    # Thanks to https://bugs.python.org/issue16487 is not possible today to easily use memory pem
    # in SSL context. We need to write it to a temporary file before
    ssl_cert_chain(ssl_context, use_ssl_cert, use_ssl_key, use_ssl_passphrase)
    return ssl_context


class ListenQueue(threading.Thread):
    def __init__(
        self,
        config: Dict,
        injector_config,
        logger,
        callback,
    ) -> None:
        threading.Thread.__init__(self)
        self.pika_credentials = None
        self.pika_parameters = None
        self.pika_connection = None
        self.channel = None

        self.callback = callback
        self.config = config
        self.logger = logger
        self.host = injector_config.connection["host"]
        self.vhost = injector_config.connection["vhost"]
        self.use_ssl = injector_config.connection["use_ssl"]
        self.port = injector_config.connection["port"]
        self.user = injector_config.connection["user"]
        self.password = injector_config.connection["pass"]
        self.queue_name = injector_config.listen
        self.exit_event = threading.Event()
        self.thread = None

    # noinspection PyUnusedLocal
    def _process_message(self, channel, method, properties, body) -> None:
        """process a message from the rabbit queue

        :param channel: channel instance
        :type channel: callable
        :param method: message methods
        :type method: callable
        :param properties: unused
        :type properties: str
        :param body: message body (data)
        :type body: str or bytes or bytearray
        """
        json_data = json.loads(body)
        # Message should be ack before processing as we don't own the processing
        # Not ACK the message here may lead to infinite re-deliver if the connector is broken
        # Also ACK, will not have any impact on the blocking aspect of the following functions
        channel.basic_ack(delivery_tag=method.delivery_tag)
        self.thread = threading.Thread(target=self._data_handler, args=[json_data])
        self.thread.start()

    def _data_handler(self, json_data) -> None:
        self.callback(json_data)

    def run(self) -> None:
        self.logger.info("Starting ListenQueue thread")
        while not self.exit_event.is_set():
            try:
                self.logger.info("ListenQueue connecting to RabbitMQ.")
                # Connect the broker
                self.pika_credentials = pika.PlainCredentials(self.user, self.password)
                self.pika_parameters = pika.ConnectionParameters(
                    host=self.host,
                    port=self.port,
                    virtual_host=self.vhost,
                    credentials=self.pika_credentials,
                    ssl_options=(
                        pika.SSLOptions(create_mq_ssl_context(self.config), self.host)
                        if self.use_ssl
                        else None
                    ),
                )
                self.pika_connection = pika.BlockingConnection(self.pika_parameters)
                self.channel = self.pika_connection.channel()
                try:
                    # confirm_delivery is only for cluster mode rabbitMQ
                    # when not in cluster mode this line raise an exception
                    self.channel.confirm_delivery()
                except Exception as err:  # pylint: disable=broad-except
                    self.logger.error(str(err))
                self.channel.basic_qos(prefetch_count=1)
                assert self.channel is not None
                self.channel.basic_consume(
                    queue=self.queue_name, on_message_callback=self._process_message
                )
                self.channel.start_consuming()
            except Exception:  # pylint: disable=broad-except
                try:
                    self.pika_connection.close()
                except Exception as errInException:
                    self.logger.error(str(errInException))
                traceback.print_exc()
                # Wait some time and then retry ListenQueue again.
                time.sleep(10)

    def stop(self):
        self.logger.info("Preparing ListenQueue for clean shutdown")
        self.exit_event.set()
        self.pika_connection.close()
        if self.thread:
            self.thread.join()


class PingAlive(utils.PingAlive):
    pass


### DEPRECATED
class OpenBASConfigHelper:
    def __init__(self, base_path, variables: Dict):
        self.__config_obj = Configuration(
            config_hints=variables,
            config_file_path=os.path.join(
                os.path.dirname(os.path.abspath(base_path)), "config.yml"
            ),
        )

    def get_conf(self, variable, is_number=None, default=None, required=None):
        result = None
        try:
            result = (
                self.__config_obj.get(variable)
                if (self.__config_obj.get(variable) is not None)
                else default
            )
        except ConfigurationError:
            result = default
        finally:
            if result is None and default is None and required:
                raise ValueError(
                    f"Could not find required key {variable} with no available default."
                )
            return result

    def to_configuration(self):
        return self.__config_obj


### DEPRECATED
class OpenBASCollectorHelper:
    def __init__(
        self,
        config: OpenBASConfigHelper,
        icon,
        security_platform_type=None,
        connect_run_and_terminate: bool = False,
    ) -> None:
        config_obj = config.to_configuration()
        # ensure the icon path is set in config
        config_obj.set("collector_icon_filepath", icon)
        # override the platform in config if passed this way
        if security_platform_type is not None:
            config_obj.set("collector_platform", security_platform_type)

        self.__daemon = CollectorDaemon(
            configuration=config_obj,
            callback=None,
        )

        self.__daemon.logger.warning(
            f"DEPRECATED: this collector should be migrated to use {CollectorDaemon}."
        )

        self.logger_class = utils.logger(
            config.get_conf("collector_log_level", default="error").upper(),
            config.get_conf("collector_json_logging", default=True),
        )
        self.collector_logger = self.logger_class(config.get_conf("collector_name"))
        self.api = self.__daemon.api
        self.config_helper = config
        self.config = {
            "collector_id": config_obj.get("collector_id"),
            "collector_name": config_obj.get("collector_name"),
            "collector_type": config_obj.get("collector_type"),
            "collector_period": config_obj.get("collector_period"),
        }

    def schedule(self, message_callback, delay):
        # backwards compatibility: when older style call sets delay
        # and no config exists,
        if self.__daemon._configuration.get("collector_period") is None:
            self.__daemon._configuration.set("collector_period", delay)
        self.__daemon.set_callback(message_callback)
        self.__daemon.start()


class OpenBASInjectorHelper:
    def __init__(self, config: OpenBASConfigHelper, icon) -> None:
        self.api = OpenBAS(
            url=config.get_conf("openbas_url"),
            token=config.get_conf("openbas_token"),
        )
        # Get the mq configuration from api
        self.config = {
            "injector_id": config.get_conf("injector_id"),
            "injector_name": config.get_conf("injector_name"),
            "injector_type": config.get_conf("injector_type"),
            "injector_contracts": config.get_conf("injector_contracts"),
            "injector_custom_contracts": config.get_conf(
                "injector_custom_contracts", default=False
            ),
            "injector_category": config.get_conf("injector_category", default=None),
            "injector_executor_commands": config.get_conf(
                "injector_executor_commands", default=None
            ),
            "injector_executor_clear_commands": config.get_conf(
                "injector_executor_clear_commands", default=None
            ),
        }

        self.logger_class = utils.logger(
            config.get_conf("injector_log_level", default="error").upper(),
            config.get_conf("injector_json_logging", default=True),
        )
        self.injector_logger = self.logger_class(config.get_conf("injector_name"))

        icon_name = config.get_conf("injector_type") + ".png"
        injector_icon = (icon_name, icon, "image/png")
        self.injector_config = self.api.injector.create(self.config, injector_icon)
        self.connect_run_and_terminate = False
        self.scheduler = sched.scheduler(time.time, time.sleep)
        # Start ping thread
        if not self.connect_run_and_terminate:
            self.ping = PingAlive(
                self.api, self.config, self.injector_logger, "injector"
            )
            self.ping.start()
        self.listen_queue = None

    def listen(self, message_callback: Callable[[Dict], None]) -> None:
        self.listen_queue = ListenQueue(
            self.config, self.injector_config, self.injector_logger, message_callback
        )
        self.listen_queue.start()


class OpenBASDetectionHelper:
    def __init__(self, logger, relevant_signatures_types) -> None:
        self.logger = logger
        self.relevant_signatures_types = relevant_signatures_types

    def match_alert_element_fuzzy(self, signature_value, alert_values, fuzzy_scoring):
        for alert_value in alert_values:
            self.logger.info(
                "Comparing alert value (" + alert_value + ", " + signature_value + ")"
            )
            ratio = fuzz.ratio(alert_value, signature_value)
            if ratio > fuzzy_scoring:
                self.logger.info("MATCHING! (score: " + str(ratio) + ")")
                return True
        return False

    def match_alert_elements(self, signatures, alert_data):
        return self._match_alert_elements_original(
            signatures, alert_data
        ) or self._match_alert_elements_for_command_line(signatures, alert_data)

    def _match_alert_elements_original(self, signatures, alert_data):
        # Example for alert_data
        # {"process_name": {"list": ["xx", "yy"], "fuzzy": 90}}
        relevant_signatures = [
            s for s in signatures if s["type"] in self.relevant_signatures_types
        ]

        # Matching logics
        signatures_number = len(relevant_signatures)
        matching_number = 0
        for signature in relevant_signatures:
            alert_data_for_signature = alert_data[signature["type"]]
            signature_result = False
            if alert_data_for_signature["type"] == "fuzzy":
                signature_result = self.match_alert_element_fuzzy(
                    signature["value"],
                    alert_data_for_signature["data"],
                    alert_data_for_signature["score"],
                )
            elif alert_data_for_signature["type"] == "simple":
                signature_result = signature["value"] in str(
                    alert_data_for_signature["data"]
                )

            if signature_result:
                matching_number = matching_number + 1

        if signatures_number == matching_number:
            return True
        return False

    def _match_alert_elements_for_command_line(self, signatures, alert_data):
        command_line_signatures = [
            signature
            for signature in signatures
            if signature.get("type") == "command_line"
        ]
        if len(command_line_signatures) == 0:
            return False
        key_types = ["command_line", "process_name", "file_name"]
        alert_datas = [alert_data.get(key) for key in key_types if key in alert_data]
        for signature in command_line_signatures:
            signature_result = False
            signature_value = self._decode_value(signature["value"]).strip().lower()
            for alert_data in alert_datas:
                trimmed_lowered_datas = [s.strip().lower() for s in alert_data["data"]]
                signature_result = any(
                    data in signature_value for data in trimmed_lowered_datas
                )
            if signature_result:
                return True
        return False

    def _decode_value(self, signature_value):
        if _is_base64_encoded(signature_value):
            try:
                decoded_bytes = base64.b64decode(signature_value)
                decoded_str = decoded_bytes.decode("utf-8")
                return decoded_str
            except Exception as e:
                self.logger.error(str(e))
        else:
            return signature_value


def _is_base64_encoded(str_maybe_base64):
    # Check if the length is a multiple of 4 and matches the Base64 character set
    base64_pattern = re.compile(r"^[A-Za-z0-9+/]*={0,2}$")
    return len(str_maybe_base64) % 4 == 0 and bool(
        base64_pattern.match(str_maybe_base64)
    )



from typing import Any, Dict

from pyobas import exceptions as exc
from pyobas.base import RESTManager, RESTObject
from pyobas.mixins import CreateMixin, ListMixin, UpdateMixin
from pyobas.utils import RequiredOptional


class User(RESTObject):
    _id_attr = "user_id"


class UserManager(CreateMixin, ListMixin, UpdateMixin, RESTManager):
    _path = "/players"
    _obj_cls = User
    _create_attrs = RequiredOptional(
        required=("user_email",),
        optional=(
            "user_firstname",
            "user_lastname",
            "user_organization",
            "user_country",
            "user_tags",
        ),
    )

    @exc.on_http_error(exc.OpenBASUpdateError)
    def upsert(self, user: Dict[str, Any], **kwargs: Any) -> Dict[str, Any]:
        path = f"{self.path}/upsert"
        result = self.openbas.http_post(path, post_data=user, **kwargs)
        return result



from typing import Any, Dict

from pyobas import exceptions as exc
from pyobas.base import RESTManager, RESTObject


class Endpoint(RESTObject):
    _id_attr = "asset_id"


class EndpointManager(RESTManager):
    _path = "/endpoints"
    _obj_cls = Endpoint

    @exc.on_http_error(exc.OpenBASUpdateError)
    def get(self, asset_id: str, **kwargs: Any) -> Dict[str, Any]:
        path = f"{self.path}/" + asset_id
        result = self.openbas.http_get(path, **kwargs)
        return result



from pyobas.base import RESTManager, RESTObject
from pyobas.mixins import CreateMixin, GetMixin, ListMixin, UpdateMixin
from pyobas.utils import RequiredOptional


class Injector(RESTObject):
    pass


class InjectorManager(GetMixin, ListMixin, CreateMixin, UpdateMixin, RESTManager):
    _path = "/injectors"
    _obj_cls = Injector
    _create_attrs = RequiredOptional(
        required=(
            "injector_id",
            "injector_name",
            "injector_type",
            "injector_contracts",
        ),
        optional=(
            "injector_custom_contracts",
            "injector_category",
            "injector_executor_commands",
            "injector_executor_clear_commands",
        ),
    )
